"""ods works with datamatrices object. This utilities converts
them to an pandas DataFrame for ease of use."""

from __future__ import annotations

import pandas as pd
import numpy as np

import odsbox.proto.ods_pb2 as ods
from odsbox.model_cache import ModelCache
from odsbox.asam_time import to_pd_timestamp


def unknown_array_values(
    unknown_array: ods.DataMatrix.Column.UnknownArray,
    date_as_timestamp: bool = False,
) -> list:
    """
    Get the values of an UnknownArray as list

    :param ods.DataMatrix.Column.UnknownArray unknown_array: ASAM ODS unknown array to transport array of any
    :param bool date_as_timestamp: columns of type DT_DATE or DS_DATE are returned as string.
                                   If this is set to True the strings are converted to pandas Timestamp.

    :raises ValueError: If standard is extended by new types
    :return list | np.array: list containing the values of the Unknown array.
    """
    if unknown_array.WhichOneof("UnknownOneOf") is None:
        return []

    if unknown_array.HasField("string_array"):
        if date_as_timestamp and ods.DT_DATE == unknown_array.data_type:
            return list(map(lambda x: to_pd_timestamp(x), unknown_array.string_array.values))
        return list(unknown_array.string_array.values)
    if unknown_array.HasField("long_array"):
        return list(unknown_array.long_array.values)
    if unknown_array.HasField("float_array"):
        if ods.DT_COMPLEX == unknown_array.data_type:
            return np.array(unknown_array.float_array.values, dtype=np.float32).view(np.complex64)
        return list(unknown_array.float_array.values)
    if unknown_array.HasField("boolean_array"):
        return list(unknown_array.boolean_array.values)
    if unknown_array.HasField("byte_array"):
        return list(unknown_array.byte_array.values)
    if unknown_array.HasField("double_array"):
        if ods.DT_DCOMPLEX == unknown_array.data_type:
            return np.array(unknown_array.double_array.values, dtype=np.float64).view(np.complex128)
        return list(unknown_array.double_array.values)
    if unknown_array.HasField("longlong_array"):
        return list(unknown_array.longlong_array.values)
    if unknown_array.HasField("bytestr_array"):
        return list(unknown_array.bytestr_array.values)

    raise ValueError(f"DataType {unknown_array.WhichOneof('UnknownOneOf')} not handled in python code!")


def __adjust_enums(
    model_cache: ModelCache | None, enumeration: ods.Model.Enumeration | None, values: list[int] | None
) -> list[int] | list[str] | None:
    if values is None or enumeration is None or model_cache is None:
        return values

    return list(map(lambda x: model_cache.enumeration_value_to_key(enumeration, x), values))


def __get_datamatrix_column_values(
    column: ods.DataMatrix.Column,
    model_cache: ModelCache | None,
    enumeration: ods.Model.Enumeration | None,
    date_as_timestamp: bool,
) -> list | None:
    if column.WhichOneof("ValuesOneOf") is None:
        return None

    if column.HasField("string_array"):
        rv = list(column.string_array.values)
        if ods.DT_EXTERNALREFERENCE == column.data_type:
            return list(zip(rv[::3], rv[1::3], rv[2::3]))
        if date_as_timestamp and ods.DT_DATE == column.data_type:
            return list(map(lambda x: to_pd_timestamp(x), rv))
        return rv
    if column.HasField("long_array"):
        if ods.DT_ENUM == column.data_type:
            return __adjust_enums(model_cache, enumeration, list(column.long_array.values))
        return list(column.long_array.values)
    if column.HasField("float_array"):
        if ods.DT_COMPLEX == column.data_type:
            return np.array(column.float_array.values, dtype=np.float32).view(np.complex64)
        return list(column.float_array.values)
    if column.HasField("boolean_array"):
        return list(column.boolean_array.values)
    if column.HasField("byte_array"):
        return list(column.byte_array.values)
    if column.HasField("double_array"):
        if ods.DT_DCOMPLEX == column.data_type:
            return np.array(column.double_array.values, dtype=np.float64).view(np.complex128)
        return list(column.double_array.values)
    if column.HasField("longlong_array"):
        return list(column.longlong_array.values)
    if column.HasField("bytestr_array"):
        return list(column.bytestr_array.values)
    # vector attributes. Look for the additional 's'
    if column.HasField("string_arrays"):
        if ods.DS_EXTERNALREFERENCE == column.data_type:
            return [
                list(zip(item.values[::3], item.values[1::3], item.values[2::3]))
                for item in column.string_arrays.values
            ]
        if date_as_timestamp and ods.DS_DATE == column.data_type:
            return [list(map(lambda x: to_pd_timestamp(x), item.values)) for item in column.string_arrays.values]
        return [list(item.values) for item in column.string_arrays.values]
    if column.HasField("long_arrays"):
        if ods.DS_ENUM == column.data_type:
            return [__adjust_enums(model_cache, enumeration, list(item.values)) for item in column.long_arrays.values]
        return [list(item.values) for item in column.long_arrays.values]
    if column.HasField("float_arrays"):
        if ods.DS_COMPLEX == column.data_type:
            return [(np.array(item.values, dtype=np.float32).view(np.complex64)) for item in column.float_arrays.values]
        return [list(item.values) for item in column.float_arrays.values]
    if column.HasField("boolean_arrays"):
        return [list(item.values) for item in column.boolean_arrays.values]
    if column.HasField("byte_arrays"):
        return [list(item.values) for item in column.byte_arrays.values]
    if column.HasField("double_arrays"):
        if ods.DS_DCOMPLEX == column.data_type:
            return [
                (np.array(item.values, dtype=np.float64).view(np.complex128)) for item in column.double_arrays.values
            ]
        return [list(item.values) for item in column.double_arrays.values]
    if column.HasField("longlong_arrays"):
        return [list(item.values) for item in column.longlong_arrays.values]
    if column.HasField("bytestr_arrays"):
        return [list(item.values) for item in column.bytestr_arrays.values]
    if column.HasField("unknown_arrays"):
        return [unknown_array_values(item, date_as_timestamp) for item in column.unknown_arrays.values]

    raise ValueError(f"DataType '{column.WhichOneof('ValuesOneOf')}' not handled!")


def __get_datamatrix_column_values_ex(
    column: ods.DataMatrix.Column,
    model_cache: ModelCache | None,
    enum_as_string: bool,
    entity: ods.Model.Entity | None,
    date_as_timestamp: bool,
) -> list:
    enumeration = None
    if (
        enum_as_string
        and entity is not None
        and model_cache is not None
        and column.data_type in [ods.DT_ENUM, ods.DS_ENUM]
    ):
        attribute = model_cache.attribute_no_throw(entity, column.name)
        if attribute is not None:
            if attribute.enumeration is not None:
                enumeration = model_cache.model().enumerations[attribute.enumeration]

    values = __get_datamatrix_column_values(column, model_cache, enumeration, date_as_timestamp)
    return_values = [] if values is None else values

    return return_values


def to_pandas(
    data_matrices: ods.DataMatrices,
    model_cache: ModelCache | None = None,
    enum_as_string: bool = False,
    date_as_timestamp: bool = False,
    name_separator: str = ".",
) -> pd.DataFrame:
    """
    Converts data in an ASAM ODS DataMatrices into a pandas DataFrame.

    :param ods.DataMatrices data_matrices: matrices to be converted.
    :param ModelCache | None model_cache: ModelCache is used to do enum conversion
    :param bool enum_as_string: columns of type DT_ENUM or DS_ENUM are returned as int values.
                                If this is set to True the model_cache is used to map the int values
                                to the corresponding string values.
    :param bool date_as_timestamp: columns of type DT_DATE or DS_DATE are returned as string.
                                   If this is set to True the strings are converted to pandas Timestamp.
    :param str name_separator: separator used to concatenate entity and attribute names to define column name.

    :return pd.DataFrame: A pandas DataFrame containing all the single matrices in a single frame. The
                          columns are named by the schema `ENTITY_NAME.ATTRIBUTE_NAME[.AGGREGATE]`.
    """
    if 0 == len(data_matrices.matrices):
        return pd.DataFrame()

    if 0 == len(data_matrices.matrices[0].columns):
        return pd.DataFrame()

    column_dict = {}

    for matrix in data_matrices.matrices:
        entity = model_cache.entity(matrix.name) if model_cache is not None else None
        for column in matrix.columns:
            aggregate_postfix = (
                ""
                if ods.AggregateEnum.AG_NONE == column.aggregate
                else name_separator + ods.AggregateEnum.Name(column.aggregate)
            )
            column_name = f"{matrix.name}{name_separator}{column.name}{aggregate_postfix}"
            # The flags are ignored here. There might be NULL in here. Check `column.is_null` for this.
            column_dict[column_name] = __get_datamatrix_column_values_ex(
                column, model_cache, enum_as_string, entity, date_as_timestamp
            )

    return pd.DataFrame(column_dict)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class NotificationTypeEnum(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    NT_ALL: _ClassVar[NotificationTypeEnum]
    NT_NEW: _ClassVar[NotificationTypeEnum]
    NT_MODIFY: _ClassVar[NotificationTypeEnum]
    NT_DELETE: _ClassVar[NotificationTypeEnum]
    NT_MODEL: _ClassVar[NotificationTypeEnum]
    NT_SECURITY: _ClassVar[NotificationTypeEnum]
NT_ALL: NotificationTypeEnum
NT_NEW: NotificationTypeEnum
NT_MODIFY: NotificationTypeEnum
NT_DELETE: NotificationTypeEnum
NT_MODEL: NotificationTypeEnum
NT_SECURITY: NotificationTypeEnum

class RegistrationRequest(_message.Message):
    __slots__ = ("mode", "entries")
    class NotificationModeEnum(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        NM_POOL: _ClassVar[RegistrationRequest.NotificationModeEnum]
        NM_PUSH: _ClassVar[RegistrationRequest.NotificationModeEnum]
    NM_POOL: RegistrationRequest.NotificationModeEnum
    NM_PUSH: RegistrationRequest.NotificationModeEnum
    class TypeElement(_message.Message):
        __slots__ = ("type", "aids")
        TYPE_FIELD_NUMBER: _ClassVar[int]
        AIDS_FIELD_NUMBER: _ClassVar[int]
        type: NotificationTypeEnum
        aids: _containers.RepeatedScalarFieldContainer[int]
        def __init__(self, type: _Optional[_Union[NotificationTypeEnum, str]] = ..., aids: _Optional[_Iterable[int]] = ...) -> None: ...
    MODE_FIELD_NUMBER: _ClassVar[int]
    ENTRIES_FIELD_NUMBER: _ClassVar[int]
    mode: RegistrationRequest.NotificationModeEnum
    entries: _containers.RepeatedCompositeFieldContainer[RegistrationRequest.TypeElement]
    def __init__(self, mode: _Optional[_Union[RegistrationRequest.NotificationModeEnum, str]] = ..., entries: _Optional[_Iterable[_Union[RegistrationRequest.TypeElement, _Mapping]]] = ...) -> None: ...

class Notification(_message.Message):
    __slots__ = ("uuid", "type", "aid", "iids", "timestamp", "user_id")
    UUID_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    AID_FIELD_NUMBER: _ClassVar[int]
    IIDS_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    USER_ID_FIELD_NUMBER: _ClassVar[int]
    uuid: str
    type: NotificationTypeEnum
    aid: int
    iids: _containers.RepeatedScalarFieldContainer[int]
    timestamp: str
    user_id: int
    def __init__(self, uuid: _Optional[str] = ..., type: _Optional[_Union[NotificationTypeEnum, str]] = ..., aid: _Optional[int] = ..., iids: _Optional[_Iterable[int]] = ..., timestamp: _Optional[str] = ..., user_id: _Optional[int] = ...) -> None: ...

class NotificationPool(_message.Message):
    __slots__ = ("events",)
    EVENTS_FIELD_NUMBER: _ClassVar[int]
    events: _containers.RepeatedCompositeFieldContainer[Notification]
    def __init__(self, events: _Optional[_Iterable[_Union[Notification, _Mapping]]] = ...) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from odsbox.proto import ods_external_data_pb2 as odsbox_dot_proto_dot_ods__external__data__pb2

GRPC_GENERATED_VERSION = '1.65.4'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.66.0'
SCHEDULED_RELEASE_DATE = 'August 6, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in odsbox/proto/ods_external_data_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class ExternalDataReaderStub(object):
    """
    EXD-API RPC service definition
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Open = channel.unary_unary(
                '/ods.external_data.ExternalDataReader/Open',
                request_serializer=odsbox_dot_proto_dot_ods__external__data__pb2.Identifier.SerializeToString,
                response_deserializer=odsbox_dot_proto_dot_ods__external__data__pb2.Handle.FromString,
                _registered_method=True)
        self.GetStructure = channel.unary_unary(
                '/ods.external_data.ExternalDataReader/GetStructure',
                request_serializer=odsbox_dot_proto_dot_ods__external__data__pb2.StructureRequest.SerializeToString,
                response_deserializer=odsbox_dot_proto_dot_ods__external__data__pb2.StructureResult.FromString,
                _registered_method=True)
        self.GetValues = channel.unary_unary(
                '/ods.external_data.ExternalDataReader/GetValues',
                request_serializer=odsbox_dot_proto_dot_ods__external__data__pb2.ValuesRequest.SerializeToString,
                response_deserializer=odsbox_dot_proto_dot_ods__external__data__pb2.ValuesResult.FromString,
                _registered_method=True)
        self.GetValuesEx = channel.unary_unary(
                '/ods.external_data.ExternalDataReader/GetValuesEx',
                request_serializer=odsbox_dot_proto_dot_ods__external__data__pb2.ValuesExRequest.SerializeToString,
                response_deserializer=odsbox_dot_proto_dot_ods__external__data__pb2.ValuesExResult.FromString,
                _registered_method=True)
        self.Close = channel.unary_unary(
                '/ods.external_data.ExternalDataReader/Close',
                request_serializer=odsbox_dot_proto_dot_ods__external__data__pb2.Handle.SerializeToString,
                response_deserializer=odsbox_dot_proto_dot_ods__external__data__pb2.Empty.FromString,
                _registered_method=True)


class ExternalDataReaderServicer(object):
    """
    EXD-API RPC service definition
    """

    def Open(self, request, context):
        """
        'OpenExternalData' opens the external data and will return a handle for further requests.
        The external data provider will keep the resource open addressed by the handle until the
        'Close' request.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStructure(self, request, context):
        """
        'GetStructure' method will allow to retrieve information about the structure of the external data.
        This method is used by an import processor to create measurements, submatrices and localcolumns
        accordingly in an ASAM ODS database.
        The ASAM ODS server will use this information to retrieve data from the external data provider
        on client request later on.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetValues(self, request, context):
        """
        'GetValues' reads external data via/like ASAM ODS data-read interface.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetValuesEx(self, request, context):
        """
        Reads external data like ASAM ODS valuematrix-read in submatrix mode.
        With this method the channel names and/or data can be retrieved.
        Channels are addressed using pattern matching their names 
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Close(self, request, context):
        """
        Close external data to free resources.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ExternalDataReaderServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Open': grpc.unary_unary_rpc_method_handler(
                    servicer.Open,
                    request_deserializer=odsbox_dot_proto_dot_ods__external__data__pb2.Identifier.FromString,
                    response_serializer=odsbox_dot_proto_dot_ods__external__data__pb2.Handle.SerializeToString,
            ),
            'GetStructure': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStructure,
                    request_deserializer=odsbox_dot_proto_dot_ods__external__data__pb2.StructureRequest.FromString,
                    response_serializer=odsbox_dot_proto_dot_ods__external__data__pb2.StructureResult.SerializeToString,
            ),
            'GetValues': grpc.unary_unary_rpc_method_handler(
                    servicer.GetValues,
                    request_deserializer=odsbox_dot_proto_dot_ods__external__data__pb2.ValuesRequest.FromString,
                    response_serializer=odsbox_dot_proto_dot_ods__external__data__pb2.ValuesResult.SerializeToString,
            ),
            'GetValuesEx': grpc.unary_unary_rpc_method_handler(
                    servicer.GetValuesEx,
                    request_deserializer=odsbox_dot_proto_dot_ods__external__data__pb2.ValuesExRequest.FromString,
                    response_serializer=odsbox_dot_proto_dot_ods__external__data__pb2.ValuesExResult.SerializeToString,
            ),
            'Close': grpc.unary_unary_rpc_method_handler(
                    servicer.Close,
                    request_deserializer=odsbox_dot_proto_dot_ods__external__data__pb2.Handle.FromString,
                    response_serializer=odsbox_dot_proto_dot_ods__external__data__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ods.external_data.ExternalDataReader', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('ods.external_data.ExternalDataReader', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class ExternalDataReader(object):
    """
    EXD-API RPC service definition
    """

    @staticmethod
    def Open(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/ods.external_data.ExternalDataReader/Open',
            odsbox_dot_proto_dot_ods__external__data__pb2.Identifier.SerializeToString,
            odsbox_dot_proto_dot_ods__external__data__pb2.Handle.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetStructure(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/ods.external_data.ExternalDataReader/GetStructure',
            odsbox_dot_proto_dot_ods__external__data__pb2.StructureRequest.SerializeToString,
            odsbox_dot_proto_dot_ods__external__data__pb2.StructureResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetValues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/ods.external_data.ExternalDataReader/GetValues',
            odsbox_dot_proto_dot_ods__external__data__pb2.ValuesRequest.SerializeToString,
            odsbox_dot_proto_dot_ods__external__data__pb2.ValuesResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetValuesEx(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/ods.external_data.ExternalDataReader/GetValuesEx',
            odsbox_dot_proto_dot_ods__external__data__pb2.ValuesExRequest.SerializeToString,
            odsbox_dot_proto_dot_ods__external__data__pb2.ValuesExResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Close(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/ods.external_data.ExternalDataReader/Close',
            odsbox_dot_proto_dot_ods__external__data__pb2.Handle.SerializeToString,
            odsbox_dot_proto_dot_ods__external__data__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)
